import io
from numbers import Number
from typing import Dict, Generator, NamedTuple, Optional, Sequence, Union
import numpy as np
import orjson
import h5py
from .utils import sanitize_array


def default(o) -> Union[list, str, None]:
    if isinstance(o, np.generic) or isinstance(o, np.ndarray):
        return o.tolist()
    if isinstance(o, complex):
        return [o.real, o.imag]
    if isinstance(o, h5py.Empty):
        return None
    if isinstance(o, bytes):
        return o.decode()
    raise TypeError


def orjson_encode(content):
    return orjson.dumps(content, default=default, option=orjson.OPT_SERIALIZE_NUMPY)


def npy_stream(array: Sequence[Number]) -> Generator[bytes, None, None]:
    """Generator to stream nD array as a .npy file.

    :param array: Data to stream
    """
    sanitized_array = sanitize_array(array)

    # Stream header
    with io.BytesIO() as buffer:
        np.lib.format.write_array_header_1_0(
            buffer, np.lib.format.header_data_from_array_1_0(sanitized_array)
        )
        header = buffer.getvalue()
    yield header

    # Taken from numpy.lib.format.write_array
    if sanitized_array.itemsize == 0:
        buffersize = 0
    else:
        # Set buffer size to 16 MiB to hide the Python loop overhead.
        buffersize = max(16 * 1024 ** 2 // sanitized_array.itemsize, 1)

    for chunk in np.nditer(
        sanitized_array,
        flags=["external_loop", "buffered", "zerosize_ok"],
        buffersize=buffersize,
        order="C",
    ):
        yield chunk.tobytes("C")


class Response(NamedTuple):
    content: Generator[bytes, None, None]
    headers: Dict[str, str]


def encode(content, encoding: Optional[str] = "json") -> Response:
    """Encode content in given encoding.

    Warning: Not all encodings supports all types of content.

    :param content:
    :param encoding:
        - `json` (default)
        - `npy`: Only nD array-like of numbers is supported
    :returns: A Response object providing:
        - encoded `content` either as bytes or a generator of bytes
        - associated `headers`
    :raises ValueError:
    """
    if encoding in ("json", None):
        return Response(
            (chunk for chunk in (orjson_encode(content),)),  # generator
            headers={"Content-Type": "application/json"},
        )
    elif encoding == "npy":
        return Response(
            npy_stream(content),
            headers={
                "Content-Type": "application/octet-stream",
                "Content-Disposition": 'attachment; filename="data.npy"',
            },
        )
    else:
        raise ValueError(f"Unsupported encoding {encoding}")



from typing import Tuple, Union
import h5py

H5pyEntity = Union[
    h5py.Dataset, h5py.Datatype, h5py.ExternalLink, h5py.Group, h5py.SoftLink
]

Selection = Union[str, int, slice, Tuple[Union[slice, int], ...], None]



"""Helpers for usage with `Flask <https://flask.palletsprojects.com/>`_"""
from h5grove.utils import PathError
from flask import abort, Blueprint, current_app, request, Response, Request
import h5py
import os
from typing import Any, Callable, Mapping, Optional

from .content import create_content, DatasetContent, ResolvedEntityContent
from .encoders import encode


__all__ = ["attr_route", "data_route", "meta_route", "URL_RULES", "BLUEPRINT"]


def make_encoded_response(content, format: Optional[str]) -> Response:
    """Prepare flask Response according to format"""
    encoded_content, headers = encode(content, format)
    response = Response(encoded_content)
    response.headers.update(headers)
    return response


def get_content(h5file: h5py.File, path: Optional[str], resolve_links: bool = True):
    """Gets contents if path is in file. Raises 404 otherwise"""
    try:
        return create_content(h5file, path, resolve_links)
    except PathError as e:
        abort(404, str(e))


def get_filename(request: Request) -> str:
    file_path = request.args.get("file")
    if file_path is None:
        raise KeyError("File argument is required")

    full_file_path = os.path.join(current_app.config["H5_BASE_DIR"], file_path)
    if not os.path.isfile(full_file_path):
        abort(404, "File not found!")

    return full_file_path


def attr_route():
    """`/attr/` endpoint handler"""
    filename = get_filename(request)
    path = request.args.get("path")
    format = request.args.get("format")

    with h5py.File(filename, mode="r") as h5file:
        content = get_content(h5file, path)
        assert isinstance(content, ResolvedEntityContent)
        return make_encoded_response(content.attributes(), format)


def data_route():
    """`/data/` endpoint handler"""
    filename = get_filename(request)
    path = request.args.get("path")
    selection = request.args.get("selection")
    format = request.args.get("format")

    with h5py.File(filename, mode="r") as h5file:
        content = get_content(h5file, path)
        assert isinstance(content, DatasetContent)
        return make_encoded_response(content.data(selection), format)


def meta_route():
    """`/meta/` endpoint handler"""
    filename = get_filename(request)
    path = request.args.get("path")
    format = request.args.get("format")
    resolve_links_arg = request.args.get("resolve_links")
    resolve_links = (
        resolve_links_arg.lower() != "false" if resolve_links_arg is not None else True
    )

    with h5py.File(filename, mode="r") as h5file:
        content = get_content(h5file, path, resolve_links)
        return make_encoded_response(content.metadata(), format)


def stats_route():
    """`/stats/` endpoint handler"""
    filename = get_filename(request)
    path = request.args.get("path")
    selection = request.args.get("selection")
    format = request.args.get("format")

    with h5py.File(filename, mode="r") as h5file:
        content = get_content(h5file, path)
        assert isinstance(content, DatasetContent)
        return make_encoded_response(content.data_stats(selection), format)


URL_RULES = {
    "/attr/": attr_route,
    "/data/": data_route,
    "/meta/": meta_route,
    "/stats/": stats_route,
}
"""Mapping of Flask URL endpoints to handlers"""


BLUEPRINT = Blueprint("h5grove", __name__)
"""Blueprint of h5grove endpoints.

It relies on `H5_BASE_DIR` being defined in the app config.
"""


def _init_blueprint(blueprint: Blueprint, url_rules: Mapping[str, Callable[[], Any]]):
    for url, view_func in url_rules.items():
        blueprint.add_url_rule(url, view_func=view_func)


_init_blueprint(BLUEPRINT, URL_RULES)



#!/usr/bin/env python
# coding: utf-8
"""Tornado-based server sample code"""
import argparse
import os.path
import tornado.web
import tornado.ioloop
from h5grove.tornadoutils import get_handlers


parser = argparse.ArgumentParser(description=__doc__)
parser.add_argument(
    "-p", "--port", type=int, default=8888, help="Port the server is listening on"
)
parser.add_argument("--ip", default="localhost", help="IP the server is listening on")
parser.add_argument(
    "--basedir", default=".", help="Base directory from which to retrieve HDF5 files"
)
options = parser.parse_args()
base_dir = os.path.abspath(options.basedir)

app = tornado.web.Application(get_handlers(base_dir, allow_origin="*"), debug=True)
app.listen(options.port, options.ip)
print(f"App is listening on {options.ip}:{options.port} serving from {base_dir}...")
tornado.ioloop.IOLoop.current().start()



"""Test tornadoutils using pytest-tornado"""
import pathlib
from typing import Callable
import pytest
from tornado.httpclient import HTTPClientError
import tornado.web

from conftest import BaseServer
import base_test

from h5grove.tornadoutils import get_handlers

# Fixtures ###


@pytest.fixture(scope="session")
def _base_dir(tmp_path_factory):
    """Fixture creating base_dir.

    Needed to share base_dir between `app` and `server` fixtures
    """
    yield tmp_path_factory.mktemp("h5grove_tornado_served").absolute()


@pytest.fixture
def app(_base_dir):
    """Fixture used by pytest-tornado"""
    return tornado.web.Application(get_handlers(_base_dir), debug=True)


class _TornadoServer(BaseServer):
    """Class of objects provided by :func:`tornado_server` fixture"""

    def __init__(self, served_dir: pathlib.Path, io_loop, http_client, base_url: str):
        super().__init__(served_dir)
        self.__io_loop = io_loop
        self.__http_client = http_client
        self.__base_url = base_url

    def __fetch(self, url: str):
        """Make a synchronous fetch of given url"""
        future = self.__http_client.fetch(self.__base_url + url)
        self.__io_loop.run_sync(lambda: future)
        return future.result()

    def _get_response(self, url: str, benchmark: Callable) -> BaseServer.Response:
        r = benchmark(lambda: self.__fetch(url))
        return BaseServer.Response(
            status=r.code, headers=list(r.headers.get_all()), content=r.body
        )

    def assert_404(self, url: str):
        with pytest.raises(HTTPClientError) as e:
            self._get_response(url, lambda f: f())
        assert e.value.code == 404


@pytest.fixture
def tornado_server(_base_dir, io_loop, http_client, base_url):
    """tornado test client-based `server` fixture.

    Provides a function to fetch endpoints from the server.
    """
    yield _TornadoServer(_base_dir, io_loop, http_client, base_url)


# Tests ###


class TestTornadoEndpoints(base_test.BaseTestEndpoints):
    """Test tornado handler enpoints using pytest-tornado"""

    @pytest.fixture
    def server(self, tornado_server):
        """Override TestEndpoints.server fixture to use pytest-tornado"""
        yield tornado_server



"""Helpers for usage with `Tornado <https://www.tornadoweb.org>`_"""
from h5grove.utils import PathError
import os
from typing import Any, Optional
import h5py
from tornado.web import RequestHandler, MissingArgumentError, HTTPError
from .content import DatasetContent, ResolvedEntityContent, create_content
from .encoders import encode


__all__ = [
    "BaseHandler",
    "AttributeHandler",
    "DataHandler",
    "MetadataHandler",
    "get_handlers",
]


class BaseHandler(RequestHandler):
    def initialize(self, base_dir, allow_origin=None) -> None:
        self.base_dir = base_dir
        self.allow_origin = allow_origin

    def get(self):
        file_path = self.get_query_argument("file")
        if file_path is None:
            raise MissingArgumentError("file")

        path = self.get_query_argument("path", None)
        format = self.get_query_argument("format", None)

        full_file_path = os.path.join(self.base_dir, file_path)
        if not os.path.isfile(full_file_path):
            raise HTTPError(status_code=404, reason="File not found!")

        with h5py.File(full_file_path, "r") as h5file:
            try:
                content = self.get_content(h5file, path)
            except PathError as e:
                raise HTTPError(status_code=404, reason=str(e))

        encoded_content_chunks, headers = encode(content, format)

        for key, value in headers.items():
            self.set_header(key, value)
        for chunk in encoded_content_chunks:
            self.write(chunk)
        self.finish()

    def get_content(self, h5file, path):
        raise NotImplementedError

    def prepare(self):
        if self.allow_origin is not None:
            self.set_header("Access-Control-Allow-Origin", self.allow_origin)

    def write_error(self, status_code: int, **kwargs: Any) -> None:
        self.prepare()
        return super().write_error(status_code, **kwargs)


class AttributeHandler(BaseHandler):
    def get_content(self, h5file, path):
        content = create_content(h5file, path)
        assert isinstance(content, ResolvedEntityContent)
        return content.attributes()


class DataHandler(BaseHandler):
    def get_content(self, h5file, path):
        selection = self.get_query_argument("selection", None)

        content = create_content(h5file, path)
        assert isinstance(content, DatasetContent)
        return content.data(selection)


class MetadataHandler(BaseHandler):
    def get_content(self, h5file, path):
        resolve_links_arg = self.get_query_argument("resolve_links", None)
        resolve_links = (
            resolve_links_arg.lower() != "false"
            if resolve_links_arg is not None
            else True
        )
        content = create_content(h5file, path, resolve_links)
        return content.metadata()


class StatisticsHandler(BaseHandler):
    def get_content(self, h5file, path):
        selection = self.get_query_argument("selection", None)

        content = create_content(h5file, path)
        assert isinstance(content, DatasetContent)
        return content.data_stats(selection)


def get_handlers(base_dir: Optional[str], allow_origin: Optional[str] = None):
    """Returns list of `Rule` arguments"""
    init_args = {"base_dir": base_dir, "allow_origin": allow_origin}
    return [
        (r"/attr/.*", AttributeHandler, init_args),
        (r"/data/.*", DataHandler, init_args),
        (r"/meta/.*", MetadataHandler, init_args),
        (r"/stats/.*", StatisticsHandler, init_args),
    ]



from typing import List
import setuptools
import sys


class Lint(setuptools.Command):
    user_options: List[str] = []
    description = "Lint"

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import subprocess

        lint_steps = [
            "black --check h5grove/ example/ test/",
            "flake8 -v",
            "mypy h5grove/ example/ test/",
        ]

        for step in lint_steps:
            cmd = step.split()[0]
            print(f"Running {cmd}...")
            errno = subprocess.call([sys.executable, "-m", *step.split()])
            if errno != 0:
                raise SystemExit(errno)
            print(f"{cmd} check passed !")


if __name__ == "__main__":
    setuptools.setup(cmdclass={"lint": Lint})



"""Benchmark data requests with server apps in example/ folder"""
import pathlib
from typing import Generator
from urllib.parse import urlencode
import h5py
import numpy as np
import pytest

# Benchmark conditions
BENCHMARK_FORMAT = "json", "npy"
BENCHMARKS = dict(
    (f"/{size}square_{dtype}", (size, dtype))
    for dtype in ("float32",)
    for size in (64, 128, 256, 512, 1024)
)


@pytest.fixture(scope="module")
def h5filepath(subprocess_server) -> Generator[pathlib.Path, None, None]:
    """Fixture providing HDF5 file for benchmark.

    Provides pathlib.Path
    """
    filepath = subprocess_server.served_directory / "test.h5"

    with h5py.File(filepath, mode="w") as h5file:
        for name, (size, dtype) in BENCHMARKS.items():
            h5file[name] = np.random.random((size, size)).astype(dtype)

    yield filepath


@pytest.mark.parametrize("format", BENCHMARK_FORMAT)
@pytest.mark.parametrize("h5path", tuple(BENCHMARKS.keys()))
def test_benchmark_data(h5filepath, subprocess_server, benchmark, h5path, format):
    """/data/ benchmark data access"""
    subprocess_server.get(
        f"/data/?{urlencode({'file': h5filepath.name, 'path': h5path, 'format': format})}",
        benchmark,
    )
